#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;

use iANT::Cfg;
require "iANT/Utils.pl";

use MIME::Base64;

my $O_PARAM;
my @A_OUTPUT_FIELDS = ();
my @A_INDEX_FIELDS  = ();

MAIN:
{

    my $create_index = 0;

    $O_PARAM = New ParamParser('GETOPTLONG', ("parser_cfg=s", "store_dir=s", "query=s", "verbose", "force","escape_special_characters"));
    $O_PARAM->SetUsage(my $usage = sub {&Usage();});


    $O_PARAM->AssertDefined('store_dir');

    my $store_dir_string = $O_PARAM->Get('store_dir');
    if ($store_dir_string =~ /[\*\?]/)
    {
        $store_dir_string = `echo $store_dir_string`;
        chomp $store_dir_string;
    }
    my @a_store_dir = split(/[,\s]/, $store_dir_string);


    $O_PARAM->AssertFileExists('parser_cfg');
    my $parser_cfg = $O_PARAM->Get('parser_cfg');
    $O_PARAM->Update($parser_cfg, 'A');
    my ($ra_fields_to_parse, $rh_docid) = &GetLuceneFields4Xml($O_PARAM);
    &GetOutputFields($ra_fields_to_parse);
    &GetIndexFields($ra_fields_to_parse);

    $O_PARAM->AssertDefined('query');

    my $analyzer_mod = $O_PARAM->Get('analyzer');
    my $analyzer     = new $analyzer_mod;

    $O_PARAM->SetBehaviour('use_substitution_table');
    $O_PARAM = &iANT::Cfg::Load($O_PARAM);

    my $unix_find = $O_PARAM->Get('unix_find');
    my $unix_gzip = $O_PARAM->Get('unix_gzip');
    my $unix_cat  = $O_PARAM->Get('unix_cat');

    my $query_string = $O_PARAM->Get('query');
    $query_string = &EscapeLuceneSpecialCharaters($query_string, 'no') if ($O_PARAM->IsDefined('escape_special_characters'));

    foreach my $store_dir (@a_store_dir)
    {
        if (!-d $store_dir)
        {
            print STDERR "ERROR:  $store_dir is not a directory\n";
            next;
        }


        
        my $store    = Lucene::Store::FSDirectory->getDirectory($store_dir, 0);
        my $searcher = new Lucene::Search::IndexSearcher($store);
        my $parser   = new Lucene::QueryParser($O_PARAM->Get('default_search'), $analyzer);
        my %h_seen   = ();

        # Substitute spaces with '__' otherwise the tests will fail. There is probably a better way
        # to handle queries with "word1 word2 ..." but I have no other idea on the moment
        $query_string =~ s/ and / & /ig;
        $query_string =~ s/ or / \| /ig;
        my $tmp_query = $query_string;
        while ($tmp_query =~ /:"(.*?)"/g)
        {
            my $tmp_str = $1;
            my $mask    = $1;
            $tmp_str      =~ s/\s+/__/g;
            $query_string =~ s/\Q$mask\E/$tmp_str/;
        }

        my $ra_docids = &GetDocids($query_string, $store, $searcher, $parser);
        my $count = $#{$ra_docids} + 1 ;
        
        if ($count > 0)
        {
            my $execute = 'y';

            print STDOUT "INFO: Will delete $count entries in $store_dir. Execute ? [y/N]\n" if (!$O_PARAM->IsDefined('force'));
            $execute = <STDIN> if (!$O_PARAM->IsDefined('force'));
            chomp $execute;

            if ($execute =~ /y/i)
            {
                foreach my $docid_value (@{$ra_docids})
                {
                    if (defined $docid_value && ($docid_value ne ''))
                    {
                        my $reader = Lucene::Index::IndexReader->open($store);
                        $docid_value = lc($docid_value) if ($rh_docid->{mode} !~ /keyword/i);
		                print STDOUT "INFO: Delete entries with term " . $rh_docid->{alias} . " '$docid_value'\n" if ($O_PARAM->IsDefined('verbose'));
                        my $term = new Lucene::Index::Term($rh_docid->{alias}, $docid_value);
                        $reader->deleteDocuments($term);
                        $reader->close();
                        undef $reader;
                    }
                }
                print STDOUT "INFO: $count entries deleted in $store_dir\n" if ($O_PARAM->IsDefined('verbose'));
            }
            else
            {
                print STDOUT "INFO: Deletion cancelled for $store_dir\n\n" if ($O_PARAM->IsDefined('verbose'));
            }   
       }
       else
       {
        print STDOUT "INFO: No match in $store_dir\n\n" if ($O_PARAM->IsDefined('verbose'));
       }
        my $writer = new Lucene::Index::IndexWriter($store, $analyzer, $create_index);
        $writer->optimize();
    }    
    
    

}


sub GetDocids
{
    my ($query_string, $store, $searcher, $parser) = @_;

    
    my $ra_docids = [];
    my $tmp_lucene_query = '';
    
    my $rh_seen = {};
    my $or = 0;
    $or = 1 if ( $query_string =~ /\|/);
    
    my ($lucene_query, $perl_query) = $query_string =~ /^(\s*\S+:\S+\s*)*(.*)$/;
    while ($perl_query =~ s/(\s*[&\|]\s+\S+:\S+\s*)//g)
    {
        $tmp_lucene_query .= $1;
    }
    if ($tmp_lucene_query ne '')
    {
        $lucene_query .= $tmp_lucene_query;
    }

    if ($lucene_query ne '')
    {
        $lucene_query =~ s/\s+&\s+/ AND /g;

        $lucene_query =~ s/__/ /g;
        my $query    = $parser->parse($lucene_query);
        my $hits     = $searcher->search($query);
        my $num_hits = $hits->length();
        $perl_query =~ s/^\s*\S\s+//;
        $perl_query =~ s/:/ =~ /g;
        for (my $i = 0; $i < $num_hits; $i++)
        {
            my $document  = $hits->doc($i);
            my $docid_value = &GetDocument_docid($document);
            next if ($rh_seen->{$docid_value} and $or);
            
            if ($perl_query eq '')
            {
              push (@{$ra_docids},$docid_value) if (defined $docid_value && ($docid_value ne ''));
              $rh_seen->{$docid_value} = 1;
            }
            else
            {
                $perl_query =~ s/__/ /g;
                my $lucene_query_string = $perl_query;

                my @a_filters = split(/[&\|!]+/, $perl_query);
                foreach my $filter (@a_filters)
                {
                    if ($filter =~ /(\S+)\s*([><=~eqn!]+)\s*(\S+)\s*/)
                    {
                        my $field    = $1;
                        my $operator = $2;
                        my $value    = $3;
                        $field =~ s/\(|\)//g;

                        my $tested_value = $document->get($field);
                        $docid_value = &GetDocument_docid($document);
                        next if ($rh_seen->{$docid_value} and $or);

                        if (($operator eq '=~') || ($operator eq '!~') || ($operator eq 'eq') || ($operator eq 'ne'))
                        {
                            $tested_value =~ s/"/\\"/g;
                            $tested_value = '"' . $tested_value . '"';
                        }

                        my $new_filter = $filter;
                        $new_filter =~ s/$field/$tested_value/;

                        $lucene_query_string =~ s/$filter/$new_filter/;
                    }
                }

                $docid_value = &GetDocument_docid($document);
                
                next if ($rh_seen->{$docid_value} and $or);
                $lucene_query_string =~ s/\s+\&{1}\s+/ \&\& /g;
                $lucene_query_string =~ s/\s+\|{1}\s+/ \|\| /g;

                if (eval($lucene_query_string))
                {
                    push (@{$ra_docids},$docid_value) if (defined $docid_value && ($docid_value ne ''));
                    $rh_seen->{$docid_value} = 1;
                }
            }
        }

    }
    else
    {
        my $reader = Lucene::Index::IndexReader->open($store);
        my $lucene_query_string = $perl_query;

        my @a_documents = ();
        my $num_docs    = $reader->numDocs();

        for (my $i = 0; $i < $num_docs; $i++)
        {
            my $document  = $reader->document($i);
            my $docid_value = &GetDocument_docid($document);
            next if ($rh_seen->{$docid_value} and $or);
            my $lucene_query_string = $query_string;

            my @a_filters = split(/[&\|]+/, $query_string);
            foreach my $filter (@a_filters)
            {
                if ($filter =~ /(\S+)\s*([><=~eqn!]+)\s*(\S+)\s*/)
                {
                    my $field        = $1;
                    my $operator     = $2;
                    my $value        = $3;
                    my $tested_value = $document->get($field);
                    if (($operator eq '=~') || ($operator eq '!~') || ($operator eq 'eq') || ($operator eq 'ne'))
                    {
                        $tested_value =~ s/"/\\"/g;
                        $tested_value = '"' . $tested_value . '"';
                    }

                    my $new_filter = $filter;
                    $new_filter          =~ s/$field/$tested_value/;
                    $lucene_query_string =~ s/$filter/$new_filter/;
                }
            }

            $lucene_query_string =~ s/\s+\&{1}\s+/ \&\& /g;
            $lucene_query_string =~ s/\s+\|{1}\s+/ \|\| /g;
            if (eval($lucene_query_string))
            {
                push (@{$ra_docids},$docid_value) if (defined $docid_value && ($docid_value ne ''));
                $rh_seen->{$docid_value} = 1;
            }
        }
    }
    return $ra_docids;
}

sub GetOutputFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_OUTPUT_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unstored/i);
    }
    return;
}

sub GetIndexFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_INDEX_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unindexed/i);
    }
    return;
}

sub GetDocument_docid
{
    my ($doc) = @_;
    my $docid_field = $O_PARAM->Get('docid');
    my $docid_value = $doc->get($docid_field);
    return $docid_value;
}

sub Usage
{
    my $index_fields  = join("\n\t\t\t\t", @A_INDEX_FIELDS);
    my $default_search = $O_PARAM->Get('default_search');

    print STDOUT <<END;
   
Usage:

   $0 --parser_cfg=<file> --store_dir=<directory> --query=<string> [--verbose] [--force]
   
 Mandatory
    --parser_cfg:   parser configuration file (contains Analyzer name, fields to parse, ...)
    --store_dir :   directory where Lucene indeces are stored
    --query     :   Lucene query string (default search field: $default_search)
                    Search fields:
                                $index_fields

 Optionnal
    --force     : remove entries without waiting for a confirmation
     --escape_special_characters 
    
Description

    Remove files from a Lucene index based on a search query
   
END

}

